"""Base Error Class for LLM Agents From Scratch."""


class LLMAgentsFromScratchError(Exception):
    """Base error for all llm-agents-from-scratch exceptions."""

    pass


class LLMAgentsFromScratchWarning(Warning):
    """Base warning for all llm-agents-from-scratch warnings."""

    pass
